islist([]).
islist([_|T]) :-islist(T).

isempty([]).

del(_, [ ], [ ]).
del(X, [X|T], L) :-del(X,T,L).
del(X, [H|T1], [H|T2]) :-X \= H, del(X,T1,T2).

conc([], Y, Y).
conc([A|X], Y, [A|Z]) :- conc(X, Y, Z).

listlen([ ],0).
listlen([_|T],N):-listlen(T,N1), N is N1 + 1.

remove_last([_],[]).
remove_last([H|T],L) :- remove_last(T,TL), L = [H|TL].

atom_chars_empty(A,C) :- isempty(A) -> C=[]; atom_chars(A,C).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% http://www.nabble.com/splitting-Lists-based-on-some-delimitor-td1429769.html
% Richard O'Keefe
    split(String, "", Result) :- !,
        characters_to_strings(String, Result).
    split(String, Delimiters, Result) :-
        real_split(String, Delimiters, Result).

    characters_to_strings([], []).
    characters_to_strings([C|Cs], [[C]|Ss]) :-
        characters_to_strings(Cs, Ss).

    real_split(String, Delimiters, Result) :-
        (   append(Substring, [Delimiter|Rest], String),
            memberchk(Delimiter, Delimiters)
        ->  Result = [Substring|Substrings],
            real_split(Rest, Delimiters, Substrings)
        ;   Result = [String]
        ).
%
% split("hk.hk.kal.l.",".",R),findall(C,(member(S,R),string_to_atom(S,A),atom_chars_empty(A,C)),L) -> L is resultant list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

split_list(String, Delimiter, L) :-
    split(String,Delimiter,R),
    findall(C,(member(S,R),string_to_atom(S,A),atom_chars_empty(A,C)),L).
    
split_char_list(CharList,Delimiter,L) :-
        string_to_atom(String,CharList),
        string_to_list(String,StringList),
        string_to_atom(DS,Delimiter),
        string_to_list(DS,DSL),
        split_list(StringList,DSL,L).    

l_replace([],[],_,_).
l_replace([H],[H],_,_).
l_replace([H1,H2|T1],[H1|T2],[A1,A2],B) :- (H1 \= A1, H2 = A2), l_replace([H2|T1],T2,[A1,A2],B).
l_replace([H1,H2|T1],[H1|T2],[A1,A2],B) :- (H1 = A1, H2 \= A2), l_replace([H2|T1],T2,[A1,A2],B).
l_replace([H1,H2|T1],[H1|T2],[A1,A2],B) :- (H1 \= A1, H2 \= A2), l_replace([H2|T1],T2,[A1,A2],B).
l_replace([A1,A2|T1],[B|T2],[A1,A2],B) :- l_replace(T1,T2,[A1,A2],B).

all_wild([],_).
all_wild([H|T],Wildcard) :- member(Wildcard,H), all_wild(T,Wildcard).

prefixes(L,Wildcard,PreL) :-
    findall(
        Prefix,
        (
        member(S,L), %S is a list of chars
        split_char_list(S,Wildcard,[Prefix|_])
        ), PreL).
        
suffixes(L,Wildcard,SufL) :-
    findall(
        Suffix,
        (
        member(S,L),
        split_char_list(S,Wildcard,L2),
        last(L2,Suffix)
        ), SufL).

all_prefix(_,[]).
all_prefix(L,[H|T]) :- is_prefix(L,H), all_prefix(L,T).

is_prefix(_,[]).
is_prefix([H|T1],[H|T2]) :- is_prefix(T1,T2).

longest_sublist(L,X) :- longest_sublist_helper(L,X,[],0).
longest_sublist_helper([],X,X,_).
longest_sublist_helper([H|T],X,CurList,MaxLength) :- (listlen(H,N), N > MaxLength) -> longest_sublist_helper(T,X,H,N); longest_sublist_helper(T,X,CurList,MaxLength).

sum_prefixes([],[]).         
sum_prefixes(L, SumPreL) :-
    longest_sublist(L,LongestSL),
    all_prefix(LongestSL,L),
    SumPreL = LongestSL.

all_suffix(_,[]).
all_suffix(L,[H|T]) :- is_suffix(L,H), all_suffix(L,T).

is_suffix(_,[]).
is_suffix(L1,L2) :- last(L2, Last), last(L1, Last), remove_last(L1,T1), remove_last(L2,T2), is_suffix(T1,T2).

sum_suffixes([],[]).         
sum_suffixes(L, SumSufL) :-
    longest_sublist(L,LongestSL),
    all_suffix(LongestSL,L),
    SumSufL = LongestSL.

concat_middles(L, Wildcard, Middles) :-
    findall(
        F1,
        (
        member(SL1,L),
        split_char_list(SL1,Wildcard,SLSplit1),
        SLSplit1 = [_|F1]
        ), NoFirst),
    findall(
        F2,
        (
        member(SLSplit2,NoFirst),
        remove_last(SLSplit2,Body),
        conc_all(Body,Wildcard,F2)
        ), NoLast),
    conc_all(NoLast,Wildcard,Middles).
    
conc_all([],_,[]).
conc_all(L,Wildcard,Result) :- conc_all_helper(L,Wildcard,Result,[]).
conc_all_helper([],Wildcard,Result,CurResult) :- append(CurResult,[Wildcard],Result).
conc_all_helper([H|T],Wildcard,Result,CurResult) :- append(CurResult,[Wildcard],CR2),append(CR2,H,CR3),conc_all_helper(T,Wildcard,Result,CR3).

get_non_wilds(L,Wildcard,L_non_wild) :-
    findall(
        SL_non_wild,
        (
        member(SL,L),
        \+member(Wildcard,SL),
        SL_non_wild = SL
        ), L_non_wild).

get_wilds(L,Wildcard,L_non_wild) :-
    findall(
        SL_non_wild,
        (
        member(SL,L),
        member(Wildcard,SL),
        SL_non_wild = SL
        ), L_non_wild).

lists_equal([H|T],H) :- lists_equal_helper(T,H).
lists_equal_helper([],_).
lists_equal_helper([H|T],L) :- L = H, lists_equal_helper(T,L).
        
wilds_not_longer(L_wild,Non_wild,Wildcard) :-
    listlen(Non_wild,N),
    findall(
        Wild2,
        (
        member(Wild,L_wild),
        del(Wildcard,Wild,Wild2)
        ), L_wild2),
    not_longer(L_wild2,N).

not_longer([],_).
not_longer([H|T],N) :- listlen(H,N2), N2 =< N, not_longer(T,N).

lists_equal_wild([],[],_).
lists_equal_wild([H1|T1],[H2|T2],Wildcard) :- (H1 = H2;H1=Wildcard;H2=Wildcard),lists_equal_wild(T1,T2,Wildcard),!.

% PList - A list of patterns
% X - A possible intersection between all patterns (as a string)
nonempty_pattern(PList, X) :-
    
    findall(
        L2,
        (
        member(Chars,PList),
        atom_chars_empty(Chars,L),
        l_replace(L,L2,['.','.'],'~')      % replaces '..' with '~' for simpler parsing work
        ), PL),
    
    (
    all_wild(PL,'~') ->
    
        (
        %Case 1: Everyone contains a wildcard. Find possible intersections.
        % An intersection is the longest of all prefixes (before first wild)
        % followed by a wild, followed by the concatenation of all data in
        % between first and last wild of each string, followed by longest
        % of all suffixes
        prefixes(PL,'~',PrePL),
        suffixes(PL,'~',SufPL),
        sum_prefixes(PrePL,SumPrePL),
        sum_suffixes(SufPL,SumSufPL),
        concat_middles(PL,'~',Middles),
        append(SumPrePL,Middles,TempList1),
        append(TempList1,SumSufPL,Atom1),
        string_to_atom(X,Atom1)
        )
        
        ;
        
        (
        %Case 2: At least one char list does not contain a wildcard.
        get_non_wilds(TL,'~',TL_non_wild),
        get_wilds(TL,'~',TL_wild),
        lists_equal(TL_non_wild,Non_wild),
        wilds_not_longer(TL_wild,Non_wild,'~'),
        canmatch(TL_wild,Non_wild,'~',R),
        string_to_list(X,R)
        )
    ).

nonempty_term(TList, X) :-
    
    findall(
        L,
        (
        member(Chars,TList),
        atom_chars_empty(Chars,L)
        ), TL),
    
    (
    all_wild(TL,'*') ->
    
        (
        %Case 1: Everyone contains a wildcard. Find possible intersections.
        % An intersection is the longest of all prefixes (before first wild)
        % followed by a wild, followed by the concatenation of all data in
        % between first and last wild of each string, followed by longest
        % of all suffixes
        prefixes(TL,'*',PreTL),
        suffixes(TL,'*',SufTL),
        sum_prefixes(PreTL,SumPreTL),
        sum_suffixes(SufTL,SumSufTL),
        concat_middles(TL,'*',Middles),
        append(SumPreTL,Middles,TempList1),
        append(TempList1,SumSufTL,Atom1),
        string_to_atom(X,Atom1)
        )
        
        ;
        
        (
        %Case 2: At least one char list does not contain a wildcard.
        get_non_wilds(TL,'*',TL_non_wild),
        get_wilds(TL,'*',TL_wild),
        lists_equal(TL_non_wild,Non_wild),
        wilds_not_longer(TL_wild,Non_wild,'*'),
        canmatch(TL_wild,Non_wild,'*',R),
        string_to_list(X,R)
        )
    ).

canmatch(LW,NW,Wildcard,R) :-
    listlen(NW,N),
    padall(LW,N,Wildcard,LP),
    trymatch([NW|LP],Wildcard,R),!. % Not sure why, but multiple wilds in a row (e.g., "he***lo") cause repeats under pad

padall([],_,_,[]).
padall([H1|T1],N,Wildcard,[H2|T2]) :-
    pad(H1,N,Wildcard,H2),
    padall(T1,N,Wildcard,T2).
    
pad([],0,_,[]).
pad([X|T1],N,Wildcard,[X|T2]) :-
    N > 0,
    M is N-1,
    pad(T1,M,Wildcard,T2).
pad([Wildcard|T1],N,Wildcard,T2) :- pad(T1,N,Wildcard,T2).
pad([Wildcard|T1],N,Wildcard,[Wildcard,Wildcard|T2]) :-
    N > 1,
    M is N-2,
    pad([Wildcard|T1],M,Wildcard,T2).
    
%trymatch([],_,[]).
%trymatch([[]|T],_,[]) :- trymatch(T,_,[]).
trymatch(L,_,[]) :-
    findall(H,member([H|_],L),CharL),
    isempty(CharL).
trymatch(L,Wildcard,[RH|RT]) :-
    findall(H,member([H|_],L),CharL),
    \+isempty(CharL) ->
        (trymatch_elements(CharL,Wildcard,Wildcard,RH),
        findall(T,member([_|T],L),NewL),
        trymatch(NewL,Wildcard,RT)).

trymatch_elements([],_,R,R).
trymatch_elements([H|T],Wildcard,Wildcard,R) :- H \= Wildcard, trymatch_elements(T,Wildcard,H,R).
trymatch_elements([H|T],Wildcard,CurChar,R) :- CurChar \= Wildcard, H = CurChar, trymatch_elements(T,Wildcard,CurChar,R).
trymatch_elements([Wildcard|T],Wildcard,CurChar,R) :- trymatch_elements(T,Wildcard,CurChar,R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% intersect_term ([[VarName,String],...],[IntersectingVars...],WitnessString)
intersect_term(VSL,IntersectingVars,Witness) :-
    (
    findall(S,member([_,S],VSL),SL),
    findall(V,member([V,_],VSL),IntersectingVars),
    nonempty_term(SL,Witness)
    )
    ;
    (
    listlen(VSL,N),
    N1 is N-1,
    N1 > 1,
    comb(N1,VSL,NewVSL),
    intersect_term(NewVSL,IntersectingVars,Witness)
    ).

intersect_comp(VL,IVList,NonIV) :-
    (
    findall(V,member(V,VL),NonIV),
    \+member(NonIV,IVList)
    )
    ;
    (
    listlen(VL,N),
    N1 is N-1,
    N1 > 1,
    comb(N1,VL,NewVL),
    intersect_comp(NewVL,IVList,NonIV)
    ).
    
% Combination (http://kti.mff.cuni.cz/~bartak/prolog/combinatorics.html)
% Returns every possible ordered combination of length N
comb(0,_,[]).
comb(N,[X|T],[X|Comb]):-N>0,N1 is N-1,comb(N1,T,Comb).
comb(N,[_|T],Comb):-N>0,comb(N,T,Comb).
