// Here is where I put all the constraint generation code. Note that
// these methods are all static and public; this code file exists
// primarily to reduce bloat in the Analysis.java file.

package checkpolicy;

import checkpolicy.prolog.PrologFile;
import java.util.*;

import checkpolicy.sat.*;
import checkpolicy.util.*;

import com.sun.org.apache.bcel.internal.Repository;

import jpl.Term;
import spox.policy.PolicyComponent.*;
import spox.policy.pointcut.PointcutComponent.*;

public class Constraint
{

    private static final SatLeaf TRUE = new SatLeaf("1");

    private static PrologFile pFile = null;

    public static HashMap<HashSet<Pointcut>, SatNode> buildSetConstraintMap(
            HashMap<Pointcut, String> varMap,
            HashMap<Edge,HashSet<Pointcut>> edgePointcuts,
            PrologFile pFile)
    {
        // Currently only handles calls

        HashMap<Call,String> callMap = new HashMap<Call,String>();
        HashMap<Call,ArrayList<Modifier>> callModMap = new HashMap<Call,ArrayList<Modifier>>();
        HashMap<Call,ArrayList<ClassId>> callClassMap = new HashMap<Call,ArrayList<ClassId>>();
        HashMap<Call,ArrayList<MemberId>> callMemMap = new HashMap<Call,ArrayList<MemberId>>();
        HashMap<Call,ArrayList<RetType>> callRetMap = new HashMap<Call,ArrayList<RetType>>();

        Iterator<Pointcut> pIter = varMap.keySet().iterator();
        while (pIter.hasNext())
        {
            Pointcut p = pIter.next();
            if (p instanceof Call)
                callMap.put((Call)p, varMap.get(p));
        }

        // Handle calls
        // Obtain modifiers, classnames, memberids, and rettypes
        Iterator<Call> cIter = callMap.keySet().iterator();
        while (cIter.hasNext())
        {
            Call c = cIter.next();
            ArrayList<Modifier> modList = new ArrayList<Modifier>();
            ArrayList<ClassId> classList = new ArrayList<ClassId>();
            ArrayList<MemberId> memList = new ArrayList<MemberId>();
            ArrayList<RetType> retList = new ArrayList<RetType>();
            ArrayList<MemberPattern> mList = c.mp;
            for (int i = 0; i < mList.size(); i++)
            {
                MemberPattern m = mList.get(i);
                if (m instanceof Modifier)
                    modList.add((Modifier)m);
                else if (m instanceof ClassId)
                    classList.add((ClassId)m);
                else if (m instanceof MemberId)
                    memList.add((MemberId)m);
                else if (m instanceof RetType)
                    retList.add((RetType)m);
            }
            callModMap.put(c, modList);
            callClassMap.put(c, classList);
            callMemMap.put(c, memList);
            callRetMap.put(c, retList);
        }

        // Now construct queries for each of the four member types
        // Members
        ArrayList<PairTuple<String,String>> pairList = new ArrayList<PairTuple<String,String>>();
        cIter = callMap.keySet().iterator();
        while (cIter.hasNext())
        {
            Call c = cIter.next();
            ArrayList<MemberId> memList = callMemMap.get(c);
            for (int i = 0; i < memList.size(); i++)
                pairList.add(new PairTuple<String,String>("'"+callMap.get(c)+"'",memList.get(i).mname.text));
        }
        Hashtable[] result = queryIntersection(pFile, "term", pairList);
        if (result == null)
            result = new Hashtable[0];

        // Query for the complement
        Hashtable[] memResult = queryComplement(pFile,castVarMap(callMap),result);

        // Class names
        pairList = new ArrayList<PairTuple<String,String>>();
        cIter = callMap.keySet().iterator();
        while (cIter.hasNext())
        {
            Call c = cIter.next();
            ArrayList<ClassId> classList = callClassMap.get(c);
            for (int i = 0; i < classList.size(); i++)
                pairList.add(new PairTuple<String,String>("'"+callMap.get(c)+"'",classList.get(i).tname.text));
        }
        result = queryIntersection(pFile, "term", pairList);
        if (result == null)
            result = new Hashtable[0];

        // Query for the complement
        Hashtable[] classResult = queryComplement(pFile,castVarMap(callMap),result);

        // Merge the results before making the constraints
        result = concatHashtableArrays(memResult, classResult);

        // Build the return SAT constraint
        HashMap<HashSet<Pointcut>,SatNode> constraint = new
                HashMap<HashSet<Pointcut>,SatNode>();
        for (int i = 0; i < result.length; i++)
        {
            Term[] t = jpl.Util.listToTermArray((Term)result[i].get("NonIntersectingVars"));
            HashSet<Pointcut> varSet = new HashSet<Pointcut>();
            for (int j = 0; j < t.length; j++)
                varSet.add(findPointcutByVar(varMap, t[j].toString()));
            constraint.put(varSet, new SatUnaryParent("-",buildConjConstraint(varSet,varMap)));
        }

        return constraint;
    }
    
    private static Hashtable[] concatHashtableArrays(
            Hashtable[] a1,
            Hashtable[] a2)
    {
        if (a1 == null && a2 == null)
            return new Hashtable[0];
        else if (a1 == null)
            return a2;
        else if (a2 == null)
            return a1;
        
        Hashtable[] a_new = new Hashtable[a1.length+a2.length];
        for (int i = 0; i < a1.length; i++)
            a_new[i] = a1[i];
        for (int i = 0; i < a2.length; i++)
            a_new[a1.length+i] = a2[i];
        return a_new;
    }

    private static HashMap<Pointcut, String> castVarMap(
            HashMap varMap)
    {
        HashMap<Pointcut,String> newVarMap = new HashMap<Pointcut,String>();
        Iterator iter = varMap.keySet().iterator();
        while (iter.hasNext())
        {
            Pointcut p = (Pointcut)iter.next();
            newVarMap.put(p, (String)varMap.get(p));
        }
        return newVarMap;
    }

    private static Hashtable[] queryComplement(
            PrologFile pFile,
            HashMap<Pointcut,String> varMap,
            Hashtable[] iVarResult)
    {
        // Convert the intersecting var results into a string list of vars
        StringBuilder iVarList = new StringBuilder();
        for (int i = 0; i < iVarResult.length; i++)
        {
            Term[] t = jpl.Util.listToTermArray((Term)iVarResult[i].get("IntersectingVars"));
            String listString = "[" + Term.toString(t) + "]";
            // Don't add in if it's a duplicate of something we already have (To reduce work load)
            if (!iVarList.toString().contains(listString))
            {
                iVarList.append(listString);
                iVarList.append(",");
            }
        }
        if (iVarList.length() > 0)
            if (iVarList.charAt(iVarList.length()-1) == ',')
                iVarList.deleteCharAt(iVarList.length()-1);

        // Query for the complement of the intersections
        StringBuilder q = new StringBuilder();
        q.append("intersect_comp([");
        Iterator<String> sIter = varMap.values().iterator();
        while (sIter.hasNext())
        {
            q.append("'" + sIter.next() + "'");
            if (sIter.hasNext())
                q.append(",");
        }
        q.append("],[");
        q.append(iVarList);
        q.append("],NonIntersectingVars)");

        System.out.println(q.toString());
        return pFile.runQuery(q.toString(),true);
    }

    private static Hashtable[] queryIntersection(
            PrologFile pFile,
            String intersectType,
            ArrayList<PairTuple<String,String>> pairList)
    {
        StringBuilder q = new StringBuilder();
        q.append("intersect_" + intersectType + "([");
        StringBuilder tupleList = new StringBuilder();
        Iterator<PairTuple<String,String>> iter = pairList.iterator();
        while (iter.hasNext())
        {
            PairTuple<String,String> pair = iter.next();
            tupleList.append("["+pair.first+",'"+pair.second+"']");
            if (iter.hasNext())
                tupleList.append(",");
        }
        q.append(tupleList);
        q.append("],IntersectingVars,Witness)");

        System.out.println(q.toString());
        return pFile.runQuery(q.toString(), true);
    }

    private static SatNode buildConjConstraint(
            HashSet<Pointcut> varSet,
            HashMap<Pointcut,String> varMap)
    {
        Iterator<Pointcut> iter = varSet.iterator();
        if (varSet.size() > 1)
        {
            Pointcut p = iter.next();
            String var = varMap.get(p);
            HashSet<Pointcut> newVarSet = (HashSet<Pointcut>) varSet.clone();
            newVarSet.remove(p);
            return new SatBinaryParent(
                    "*",
                    new SatLeaf(var),
                    buildConjConstraint(newVarSet,varMap));
        }
        else if (varSet.size() == 1)
        {
            return new SatLeaf(varMap.get(iter.next()));
        }
        else
        {
            return TRUE;
        }
    }

    private static Pointcut findPointcutByVar(
            HashMap<Pointcut,String> varMap,
            String var)
    {
        Iterator<Pointcut> iter = varMap.keySet().iterator();
        while (iter.hasNext())
        {
            Pointcut p = iter.next();
            if (varMap.get(p).equals(var.replaceAll("'", "")))
                return p;
        }
        return null;
    }

    /**
     * Builds a mapping from every possible pointcut pair to a constraint,
     * where the constraints are in the form of SAT objects.
     * @param varMap The mapping from pointcuts to their variables.
     * @param pFile A Prolog file, for use in specialized queries.
     * @return The mapping from pairs to constraints.
     */
    public static HashMap<PairTuple<Pointcut, Pointcut>, SatNode>
            buildPairConstraintMap(HashMap<Pointcut, String> varMap,
            PrologFile pFile)
    {
        HashMap<PairTuple<Pointcut, Pointcut>, SatNode> constraintMap =
                new HashMap<PairTuple<Pointcut, Pointcut>, SatNode>();

        Constraint.pFile = pFile;

        // Scan through each possible pairing
        Iterator<Pointcut> iter1 = varMap.keySet().iterator();
        while (iter1.hasNext())
        {
            Pointcut p1 = iter1.next();
            Iterator<Pointcut> iter2 = varMap.keySet().iterator();
            while (iter2.hasNext())
            {
                Pointcut p2 = iter2.next();
                PairTuple<Pointcut,Pointcut> pair =
                        new PairTuple<Pointcut,Pointcut>(p1,p2);
                if (!constraintMap.containsKey(pair) &&
                    !constraintMap.containsKey(pair.flipPair()))
                {
                    constraintMap.put(pair, buildPairConstraint(varMap, pair));
                }
            }
        }

        return constraintMap;
    }

    /**
     * Builds a constraint from a given PairTuple of pointcuts.
     * @param varMap The mapping from pointcuts to their variable names.
     * @param pair The pair of pointcuts to consider.
     * @param pFile A Prolog file, for use in specialized query operations.
     * @return A SatNode representing a boolean sentence, describing the
     *         constraint. If no constraint is necessary, a generic
     *         leaf node of "1" (true) is returned.
     */
    private static SatNode buildPairConstraint(
            HashMap<Pointcut, String> varMap,
            PairTuple<Pointcut,Pointcut> pair)
    {

        // First get the var names and make leaf objects
        // (Yes, this means all leaves will have the same pointer)
        SatLeaf v1 = new SatLeaf(varMap.get(pair.first));
        SatLeaf v2 = new SatLeaf(varMap.get(pair.second));

        if (pair.first.equals(pair.second))
            return TRUE;

        // Check the pair against each possible combination
        if (pair.first instanceof Call)
        {
            Call p1 = (Call)pair.first;
            if (pair.second instanceof Call)
            {
                Call p2 = (Call)pair.second;
                return call_call(p1,p2,v1,v2);
            }
            else if (pair.second instanceof Execution)
            {
                Execution p2 = (Execution)pair.second;
                return call_execution(p1,p2,v1,v2);
            }
            else if (pair.second instanceof GetField)
            {
                GetField p2 = (GetField)pair.second;
                return call_getfield(p1,p2,v1,v2);
            }
            else if (pair.second instanceof SetField)
            {
                SetField p2 = (SetField)pair.second;
                return call_setfield(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgTyp)
            {
                ArgTyp p2 = (ArgTyp)pair.second;
                return call_argtyp(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgVal)
            {
                ArgVal p2 = (ArgVal)pair.second;
                return call_argval(p1,p2,v1,v2);
            }
        }

        else if (pair.first instanceof Execution)
        {
            Execution p1 = (Execution)pair.first;
            if (pair.second instanceof Call)
            {
                Call p2 = (Call)pair.second;
                return call_execution(p2,p1,v2,v1);
            }
            else if (pair.second instanceof Execution)
            {
                Execution p2 = (Execution)pair.second;
                return execution_execution(p1,p2,v1,v2);
            }
            else if (pair.second instanceof GetField)
            {
                GetField p2 = (GetField)pair.second;
                return execution_getfield(p1,p2,v1,v2);
            }
            else if (pair.second instanceof SetField)
            {
                SetField p2 = (SetField)pair.second;
                return execution_setfield(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgTyp)
            {
                ArgTyp p2 = (ArgTyp)pair.second;
                return execution_argtyp(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgVal)
            {
                ArgVal p2 = (ArgVal)pair.second;
                return execution_argval(p1,p2,v1,v2);
            }
        }

        else if (pair.first instanceof GetField)
        {
            GetField p1 = (GetField)pair.first;
            if (pair.second instanceof Call)
            {
                Call p2 = (Call)pair.second;
                return call_getfield(p2,p1,v2,v1);
            }
            else if (pair.second instanceof Execution)
            {
                Execution p2 = (Execution)pair.second;
                return execution_getfield(p2,p1,v2,v1);
            }
            else if (pair.second instanceof GetField)
            {
                GetField p2 = (GetField)pair.second;
                return getfield_getfield(p1,p2,v1,v2);
            }
            else if (pair.second instanceof SetField)
            {
                SetField p2 = (SetField)pair.second;
                return getfield_setfield(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgTyp)
            {
                ArgTyp p2 = (ArgTyp)pair.second;
                return getfield_argtyp(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgVal)
            {
                ArgVal p2 = (ArgVal)pair.second;
                return getfield_argval(p1,p2,v1,v2);
            }
        }

        else if (pair.first instanceof SetField)
        {
            SetField p1 = (SetField)pair.first;
            if (pair.second instanceof Call)
            {
                Call p2 = (Call)pair.second;
                return call_setfield(p2,p1,v2,v1);
            }
            else if (pair.second instanceof Execution)
            {
                Execution p2 = (Execution)pair.second;
                return execution_setfield(p2,p1,v2,v1);
            }
            else if (pair.second instanceof GetField)
            {
                GetField p2 = (GetField)pair.second;
                return getfield_setfield(p2,p1,v2,v1);
            }
            else if (pair.second instanceof SetField)
            {
                SetField p2 = (SetField)pair.second;
                return setfield_setfield(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgTyp)
            {
                ArgTyp p2 = (ArgTyp)pair.second;
                return setfield_argtyp(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgVal)
            {
                ArgVal p2 = (ArgVal)pair.second;
                return setfield_argval(p1,p2,v1,v2);
            }
        }

        else if (pair.first instanceof ArgTyp)
        {
            ArgTyp p1 = (ArgTyp)pair.first;
            if (pair.second instanceof Call)
            {
                Call p2 = (Call)pair.second;
                return call_argtyp(p2,p1,v2,v1);
            }
            else if (pair.second instanceof Execution)
            {
                Execution p2 = (Execution)pair.second;
                return execution_argtyp(p2,p1,v2,v1);
            }
            else if (pair.second instanceof GetField)
            {
                GetField p2 = (GetField)pair.second;
                return getfield_argtyp(p2,p1,v2,v1);
            }
            else if (pair.second instanceof SetField)
            {
                SetField p2 = (SetField)pair.second;
                return setfield_argtyp(p2,p1,v2,v1);
            }
            else if (pair.second instanceof ArgTyp)
            {
                ArgTyp p2 = (ArgTyp)pair.second;
                return argtyp_argtyp(p1,p2,v1,v2);
            }
            else if (pair.second instanceof ArgVal)
            {
                ArgVal p2 = (ArgVal)pair.second;
                return argtyp_argval(p1,p2,v1,v2);
            }
        }

        else if (pair.first instanceof ArgVal)
        {
            ArgVal p1 = (ArgVal)pair.first;
            if (pair.second instanceof Call)
            {
                Call p2 = (Call)pair.second;
                return call_argval(p2,p1,v2,v1);
            }
            else if (pair.second instanceof Execution)
            {
                Execution p2 = (Execution)pair.second;
                return execution_argval(p2,p1,v2,v1);
            }
            else if (pair.second instanceof GetField)
            {
                GetField p2 = (GetField)pair.second;
                return getfield_argval(p2,p1,v2,v1);
            }
            else if (pair.second instanceof SetField)
            {
                SetField p2 = (SetField)pair.second;
                return setfield_argval(p2,p1,v2,v1);
            }
            else if (pair.second instanceof ArgTyp)
            {
                ArgTyp p2 = (ArgTyp)pair.second;
                return argtyp_argval(p2,p1,v2,v1);
            }
            else if (pair.second instanceof ArgVal)
            {
                ArgVal p2 = (ArgVal)pair.second;
                return argval_argval(p1,p2,v1,v2);
            }
        }

        // Default case if nothing above matched
        return TRUE;
    }

    /**
     * Call vs. Call. If the calls are equivalent, then
     * the constraint is A<==>B (i.e., (A+!B)*(!A+B)). If not,
     * the constraint is !(A*B).
     * TODO: Expand to handle POSSIBLE intersection.
     */
    private static SatNode call_call(Call p1, Call p2, SatLeaf v1, SatLeaf v2)
    {
        final SatNode EQUAL = new SatBinaryParent("*",
                        new SatBinaryParent("+",v1,new SatUnaryParent("-",v2)),
                        new SatBinaryParent("+",v2,new SatUnaryParent("-",v1)));
        final SatNode NOT_A_AND_B = new SatUnaryParent("-",
                        new SatBinaryParent("*",v1,v2));
        final SatNode A_IMPLIES_B = new SatBinaryParent("+",
                new SatUnaryParent("-",v1), v2);
        final SatNode B_IMPLIES_A = new SatBinaryParent("+",
                v1, new SatUnaryParent("-",v2));
        final SatNode A_OR_B = new SatBinaryParent("+",v1,v2);
        final SatNode A_AND_B = new SatBinaryParent("*",v1,v2);
        final SatNode A_XOR_B = new SatBinaryParent("*",
                A_OR_B, NOT_A_AND_B);

        // Handle simplest matching case
        if (p1.equals(p2))
            return EQUAL;

        // Now special cases (conjunctively join sub-constraints)
        SatNode constraint = new SatLeaf("1");

        ArrayList<MemberPattern> mList1 = p1.mp;
        ArrayList<MemberPattern> mList2 = p2.mp;

        // Convert to modifiers, classnames, memberids, and rettypes
        ArrayList<Modifier> modList1 = new ArrayList<Modifier>();
        ArrayList<ClassId> classList1 = new ArrayList<ClassId>();
        ArrayList<MemberId> memList1 = new ArrayList<MemberId>();
        ArrayList<RetType> retList1 = new ArrayList<RetType>();
        for (int i = 0; i < mList1.size(); i++)
        {
            MemberPattern m = mList1.get(i);
            if (m instanceof Modifier)
                modList1.add((Modifier)m);
            else if (m instanceof ClassId)
                classList1.add((ClassId)m);
            else if (m instanceof MemberId)
                memList1.add((MemberId)m);
            else if (m instanceof RetType)
                retList1.add((RetType)m);
        }
        ArrayList<Modifier> modList2 = new ArrayList<Modifier>();
        ArrayList<ClassId> classList2 = new ArrayList<ClassId>();
        ArrayList<MemberId> memList2 = new ArrayList<MemberId>();
        ArrayList<RetType> retList2 = new ArrayList<RetType>();
        for (int i = 0; i < mList2.size(); i++)
        {
            MemberPattern m = mList2.get(i);
            if (m instanceof Modifier)
                modList2.add((Modifier)m);
            else if (m instanceof ClassId)
                classList2.add((ClassId)m);
            else if (m instanceof MemberId)
                memList2.add((MemberId)m);
            else if (m instanceof RetType)
                retList2.add((RetType)m);
        }

        // Handle modifier comparisons
        // Note that we assume possible matches unless two contradictory
        // modifiers (public and private) are found
        // TODO: Expand to include all possibly contradictory modifiers...
        SatNode curNode = constraint;
        for (int i = 0; i < modList1.size(); i++)
        {
            Modifier mod1 = modList1.get(i);
            for (int j = 0; j < modList2.size(); j++)
            {
                Modifier mod2 = modList2.get(i);
                if ((mod1.mname.text.equals("public") && mod2.mname.text.equals("private")) ||
                    (mod1.mname.text.equals("private") && mod2.mname.text.equals("public")))
                    return NOT_A_AND_B;
            }
        }

        // Handle ClassId comparisons
        // TODO: Handle this with regex ops
        for (int i = 0; i < classList1.size(); i++)
        {
            ClassId c1 = classList1.get(i);
            for (int j = 0; j < classList2.size(); j++)
            {
                ClassId c2 = classList2.get(i);
                if (!c1.tname.text.equals(c2.tname.text))
                {
                    if (!(c1.tname.text.contains("*") || c2.tname.text.contains("*")))
                    {
                        //?
                    }
                }
            }
        }

        // Handle MemberId comparisons
        // TODO: Same as above

        // Handle RetTyp comparisons
        // TODO: Same

        // Default case
        return TRUE;
    }

    /**
     * Execution vs. Execution. If the executions are equivalent, then
     * the constraint is A<==>B (i.e., (A+!B)*(!A+B)). If not,
     * the constraint is !(A*B).
     * TODO: Expand to handle POSSIBLE intersection.
     */
    private static SatNode execution_execution(Execution p1, Execution p2, SatLeaf v1, SatLeaf v2)
    {
        if (p1.equals(p2))
        {
            return new SatBinaryParent("*",
                    new SatBinaryParent("+",v1,new SatUnaryParent("-",v2)),
                    new SatBinaryParent("+",v2,new SatUnaryParent("-",v1)));
        }
        else
        {
            return new SatUnaryParent("-",
                    new SatBinaryParent("*",v1,v2));
        }
    }

    /**
     * Call vs. Execution. These two will never be simultaneously true, so !(A*B).
     */
    private static SatNode call_execution(Call p1, Execution p2, SatLeaf v1, SatLeaf v2)
    {
        return new SatUnaryParent("-",
                new SatBinaryParent("*",v1,v2));
    }

    /**
     * Call vs. GetField. These will never be simultaneously true, so !(A*B)
     */
    private static SatNode call_getfield(Call p1, GetField p2, SatLeaf v1, SatLeaf v2)
    {
        return new SatUnaryParent("-",
                new SatBinaryParent("*",v1,v2));
    }

    /**
     * Call vs. SetField. These will never be simultaneously true, so !(A*B)
     */
    private static SatNode call_setfield(Call p1, SetField p2, SatLeaf v1, SatLeaf v2)
    {
        return new SatUnaryParent("-",
                new SatBinaryParent("*",v1,v2));
    }

    /**
     * Call vs. ArgTyp. These do not affect one another, so TRUE.
     */
    private static SatNode call_argtyp(Call p1, ArgTyp p2, SatLeaf v1, SatLeaf v2)
    {
        return TRUE;
    }

    /**
     * Call vs. ArgTyp. These do not affect one another, so TRUE.
     */
    private static SatNode call_argval(Call p1, ArgVal p2, SatLeaf v1, SatLeaf v2)
    {
        return TRUE;
    }

    /**
     * Execution vs. GetField. These will never be simultaneously true, so !(A*B)
     */
    private static SatNode execution_getfield(Execution p1, GetField p2, SatLeaf v1, SatLeaf v2)
    {
        return new SatUnaryParent("-",
                new SatBinaryParent("*",v1,v2));
    }

    /**
     * Execution vs. SetField. These will never be simultaneously true, so !(A*B)
     */
    private static SatNode execution_setfield(Execution p1, SetField p2, SatLeaf v1, SatLeaf v2)
    {
        return new SatUnaryParent("-",
                new SatBinaryParent("*",v1,v2));
    }

    /**
     * Execution vs. ArgTyp. These do not affect one another, so TRUE.
     */
    private static SatNode execution_argtyp(Execution p1, ArgTyp p2, SatLeaf v1, SatLeaf v2)
    {
        return TRUE;
    }

    /**
     * Execution vs. ArgTyp. These do not affect one another, so TRUE.
     */
    private static SatNode execution_argval(Execution p1, ArgVal p2, SatLeaf v1, SatLeaf v2)
    {
        return TRUE;
    }

    /** STUB
     * GetField vs. GetField. Match up on individual parameters.
     */
    private static SatNode getfield_getfield(GetField p1, GetField p2, SatLeaf v1, SatLeaf v2)
    {
        throw new UnsupportedOperationException("Not yet implemented");
    }

    /** STUB
     * GetField vs. SetField. Match up on individual parameters
     */
    private static SatNode getfield_setfield(GetField p1, SetField p2, SatLeaf v1, SatLeaf v2)
    {
        throw new UnsupportedOperationException("Not yet implemented");
    }

    /**
     * GetField vs. ArgTyp. If the ArgTyp is looking for a parameter > 0, then
     * p2=>!p1. Otherwise they don't affect each other, so return TRUE;
     */
    private static SatNode getfield_argtyp(GetField p1, ArgTyp p2, SatLeaf v1, SatLeaf v2)
    {
        if (p2.num>0)
        {
            return new SatBinaryParent("+",
                    new SatUnaryParent("-",v1), new SatUnaryParent("-",v2));
        }
        else
            return TRUE;
    }

    /**
     * GetField vs. ArgVal. If the ArgVal is looking for a parameter > 0, then
     * p2=>!p1. Otherwise they don't affect each other, so return TRUE.
     */
    private static SatNode getfield_argval(GetField p1, ArgVal p2, SatLeaf v2, SatLeaf v1)
    {
        if (p2.num>0)
        {
            return new SatBinaryParent("+",
                    new SatUnaryParent("-",v1), new SatUnaryParent("-",v2));
        }
        else
            return TRUE;
    }

    /** STUB
     * SetField vs. SetField. Compare on parameters.
     */
    private static SatNode setfield_setfield(SetField p1, SetField p2, SatLeaf v1, SatLeaf v2)
    {
        throw new UnsupportedOperationException("Not yet implemented");
    }

    /**
     * GetField vs. ArgTyp. If the ArgTyp is looking for a parameter > 1, then
     * p2=>!p1. Otherwise they don't affect each other, so return TRUE;
     */
    private static SatNode setfield_argtyp(SetField p1, ArgTyp p2, SatLeaf v1, SatLeaf v2)
    {
        if (p2.num>1)
        {
            return new SatBinaryParent("+",
                    new SatUnaryParent("-",v1), new SatUnaryParent("-",v2));
        }
        else
            return TRUE;
    }

    /**
     * SetField vs. ArgVal. If the ArgVal is looking for a parameter > 1, then
     * p2=>!p1. Otherwise they don't affect each other, so return TRUE.
     */
    private static SatNode setfield_argval(SetField p1, ArgVal p2, SatLeaf v2, SatLeaf v1)
    {
        if (p2.num>1)
        {
            return new SatBinaryParent("+",
                    new SatUnaryParent("-",v1), new SatUnaryParent("-",v2));
        }
        else
            return TRUE;
    }

    /** STUB
     * ArgTyp vs. ArgTyp.
     */
    private static SatNode argtyp_argtyp(ArgTyp p1, ArgTyp p2, SatLeaf v1, SatLeaf v2)
    {
        throw new UnsupportedOperationException("Not yet implemented");
    }

    /**
     * ArgTyp vs. ArgVal. If each points to different arguments, there is
     * no constraint and we return "1" (true). Otherwise, we infer argument
     * types from the ArgVal and compare. If they are necessarily different
     * types, we return !(A*B).
     */
    private static SatNode argtyp_argval(ArgTyp p1, ArgVal p2, SatLeaf v2, SatLeaf v1)
    {
        if (p1.num!=p2.num)
            return TRUE;

        else if (!p1.tname.text.equals("int") && p2.valExpr instanceof IntCmp)
        {
            return new SatUnaryParent("-",
                        new SatBinaryParent("*",v1,v2));
        }

        return TRUE;
    }

    /**
     * ArgVal vs. ArgVal. If each argval points to different arguments, there
     * is no constraint and we return "1" (true). If each argval handles
     * different data types, we return !(A*B). Otherwise, we do an
     * analysis of the contents to determine the constraint.
     */
    private static SatNode argval_argval(ArgVal p1, ArgVal p2, SatLeaf v1, SatLeaf v2)
    {

        final SatNode EQUAL = new SatBinaryParent("*",
                        new SatBinaryParent("+",v1,new SatUnaryParent("-",v2)),
                        new SatBinaryParent("+",v2,new SatUnaryParent("-",v1)));
        final SatNode NOT_A_AND_B = new SatUnaryParent("-",
                        new SatBinaryParent("*",v1,v2));
        final SatNode A_IMPLIES_B = new SatBinaryParent("+",
                new SatUnaryParent("-",v1), v2);
        final SatNode B_IMPLIES_A = new SatBinaryParent("+",
                v1, new SatUnaryParent("-",v2));
        final SatNode A_OR_B = new SatBinaryParent("+",v1,v2);
        final SatNode A_AND_B = new SatBinaryParent("*",v1,v2);
        final SatNode A_XOR_B = new SatBinaryParent("*",
                A_OR_B, NOT_A_AND_B);
        
        if (p1.num != p2.num)
        {
            return TRUE;
        }

        else if (p1.valExpr instanceof IsNull)
        {
            if (!(p2.valExpr instanceof IsNull))
            {
                // A null and a non-null will never be simultaneously true; return !(A*B)
                return NOT_A_AND_B;
            }
            else
            {
                // Holds if the two are equal
                return EQUAL;
            }
        }

        else if (p1.valExpr instanceof StrEq)
        {
            if (!(p2.valExpr instanceof StrEq))
            {
                // Must be of the same type to be simultaneously true
                return NOT_A_AND_B;
            }
            else
            {
                // For now, just consider equality
                // TODO: Handle regexes
                String regex1 = ((StrEq)p1.valExpr).regex;
                String regex2 = ((StrEq)p2.valExpr).regex;
                if (regex1.equals(regex2))
                {
                    return EQUAL;
                }
                else
                {
                    return NOT_A_AND_B;
                }
            }
        }

        else if (p1.valExpr instanceof IntCmp)
        {
            if (!(p2.valExpr instanceof IntCmp))
            {
                // Must be of the same type to be simultaneously true
                return NOT_A_AND_B;
            }
            else
            {
                int n1 = ((IntCmp)p1.valExpr).n;
                int n2 = ((IntCmp)p2.valExpr).n;
                // Now handle each combination (36 in all!)
                if (p1.valExpr instanceof IntEQ &&
                    p2.valExpr instanceof IntEQ)
                {
                    if (n1==n2)
                        return EQUAL;
                    else
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntEQ &&
                         p2.valExpr instanceof IntGE)
                {
                    if (n2 <= n1)
                        return A_IMPLIES_B;
                    if (n2 > n1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntEQ &&
                         p2.valExpr instanceof IntGT)
                {
                    if (n2<n1)
                        return A_IMPLIES_B;
                    if (n2>=n1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntEQ &&
                         p2.valExpr instanceof IntLE)
                {
                    if (n2 >= n1)
                        return A_IMPLIES_B;
                    if (n2 < n1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntEQ &&
                         p2.valExpr instanceof IntLT)
                {
                    if (n2>n1)
                        return A_IMPLIES_B;
                    if (n2<=n1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntEQ &&
                         p2.valExpr instanceof IntNE)
                {
                    if (n1==n2)
                        return A_XOR_B;
                    if (n1!=n2)
                        return A_IMPLIES_B;
                }
                else if (p1.valExpr instanceof IntGE &&
                         p2.valExpr instanceof IntEQ)
                {
                    if (n1 <= n2)
                        return B_IMPLIES_A;
                    if (n1 > n2)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntGE &&
                         p2.valExpr instanceof IntGE)
                {
                    if (n1>n2)
                        return A_IMPLIES_B;
                    if (n1<n2)
                        return B_IMPLIES_A;
                    if (n1==n2)
                        return EQUAL;
                }
                else if (p1.valExpr instanceof IntGE &&
                         p2.valExpr instanceof IntGT)
                {
                    if (n1==n2+1)
                        return EQUAL;
                    if (n1<=n2)
                        return B_IMPLIES_A;
                    if (n1>n2)
                        return A_IMPLIES_B;
                }
                else if (p1.valExpr instanceof IntGE &&
                         p2.valExpr instanceof IntLE)
                {
                    if (n1<=n2)
                        return A_OR_B;
                    if (n1==n2+1)
                        return A_XOR_B;
                    if (n1>n2+1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntGE &&
                         p2.valExpr instanceof IntLT)
                {
                    if (n1<=n2-1)
                        return A_OR_B;
                    if (n1==n2)
                        return A_XOR_B;
                    if (n1>n2)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntGE &&
                         p2.valExpr instanceof IntNE)
                {
                    if (n1>n2)
                        return A_IMPLIES_B;
                    if (n1<=n2)
                        return A_OR_B;
                }
                else if (p1.valExpr instanceof IntGT &&
                         p2.valExpr instanceof IntEQ)
                {
                    if (n1<n2)
                        return B_IMPLIES_A;
                    if (n1>=n2)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntGT &&
                         p2.valExpr instanceof IntGE)
                {
                    if (n2==n1+1)
                        return EQUAL;
                    if (n2<=n1)
                        return A_IMPLIES_B;
                    if (n2>n1)
                        return B_IMPLIES_A;
                }
                else if (p1.valExpr instanceof IntGT &&
                         p2.valExpr instanceof IntGT)
                {
                    if (n1>n2)
                        return A_IMPLIES_B;
                    if (n1<n2)
                        return B_IMPLIES_A;
                    if (n1==n2)
                        return EQUAL;
                }
                else if (p1.valExpr instanceof IntGT &&
                         p2.valExpr instanceof IntLE)
                {
                    if (n2>=n1+1)
                        return A_OR_B;
                    if (n2==n1)
                        return A_XOR_B;
                    if (n2<n1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntGT &&
                         p2.valExpr instanceof IntLT)
                {
                    if (n1<=n2-2)
                        return A_OR_B;
                    if (n1==n2-1)
                        return A_XOR_B;
                    if (n1>n2-1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntGT &&
                         p2.valExpr instanceof IntNE)
                {
                    if (n1>=n2)
                        return A_IMPLIES_B;
                    if (n1<n2)
                        return A_OR_B;
                }
                else if (p1.valExpr instanceof IntLE &&
                         p2.valExpr instanceof IntEQ)
                {
                    if (n1 >= n2)
                        return B_IMPLIES_A;
                    if (n1 < n2)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntLE &&
                         p2.valExpr instanceof IntGE)
                {
                    if (n2<=n1)
                        return A_OR_B;
                    if (n2==n1+1)
                        return A_XOR_B;
                    if (n2>n1+1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntLE &&
                         p2.valExpr instanceof IntGT)
                {
                    if (n1>=n2+1)
                        return A_OR_B;
                    if (n1==n2)
                        return A_XOR_B;
                    if (n1<n2)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntLE &&
                         p2.valExpr instanceof IntLE)
                {
                    if (n1<n2)
                        return A_IMPLIES_B;
                    if (n1>n2)
                        return B_IMPLIES_A;
                    if (n1==n2)
                        return EQUAL;
                }
                else if (p1.valExpr instanceof IntLE &&
                         p2.valExpr instanceof IntLT)
                {
                    if (n1==n2-1)
                        return EQUAL;
                    if (n1>=n2)
                        return B_IMPLIES_A;
                    if (n1<n2)
                        return A_IMPLIES_B;
                }
                else if (p1.valExpr instanceof IntLE &&
                         p2.valExpr instanceof IntNE)
                {
                    if (n1<n2)
                        return A_IMPLIES_B;
                    if (n1>=n2)
                        return A_OR_B;
                }
                else if (p1.valExpr instanceof IntLT &&
                         p2.valExpr instanceof IntEQ)
                {
                    if (n1>n2)
                        return B_IMPLIES_A;
                    if (n1<=n2)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntLT &&
                         p2.valExpr instanceof IntGE)
                {
                    if (n2<=n1-1)
                        return A_OR_B;
                    if (n2==n1)
                        return A_XOR_B;
                    if (n2>n1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntLT &&
                         p2.valExpr instanceof IntGT)
                {
                    if (n2<=n1-2)
                        return A_OR_B;
                    if (n2==n1-1)
                        return A_XOR_B;
                    if (n2>n1-1)
                        return NOT_A_AND_B;
                }
                else if (p1.valExpr instanceof IntLT &&
                         p2.valExpr instanceof IntLE)
                {
                    if (n2==n1-1)
                        return EQUAL;
                    if (n2>=n1)
                        return A_IMPLIES_B;
                    if (n2<n1)
                        return B_IMPLIES_A;
                }
                else if (p1.valExpr instanceof IntLT &&
                         p2.valExpr instanceof IntLT)
                {
                    if (n1<n2)
                        return A_IMPLIES_B;
                    if (n1>n2)
                        return B_IMPLIES_A;
                    if (n1==n2)
                        return EQUAL;
                }
                else if (p1.valExpr instanceof IntLT &&
                         p2.valExpr instanceof IntNE)
                {
                    if (n1<=n2)
                        return A_IMPLIES_B;
                    if (n1>n2)
                        return A_OR_B;
                }
                else if (p1.valExpr instanceof IntNE &&
                         p2.valExpr instanceof IntEQ)
                {
                    if (n2==n1)
                        return A_XOR_B;
                    if (n2!=n1)
                        return B_IMPLIES_A;
                }
                else if (p1.valExpr instanceof IntNE &&
                         p2.valExpr instanceof IntGE)
                {
                    if (n2>n1)
                        return B_IMPLIES_A;
                    if (n2<=n1)
                        return A_OR_B;
                }
                else if (p1.valExpr instanceof IntNE &&
                         p2.valExpr instanceof IntGT)
                {
                    if (n2>=n1)
                        return B_IMPLIES_A;
                    if (n2<n1)
                        return A_OR_B;
                }
                else if (p1.valExpr instanceof IntNE &&
                         p2.valExpr instanceof IntLE)
                {
                    if (n1<n2)
                        return A_IMPLIES_B;
                    if (n1>=n2)
                        return A_OR_B;
                }
                else if (p1.valExpr instanceof IntNE &&
                         p2.valExpr instanceof IntLT)
                {
                    if (n2<=n1)
                        return B_IMPLIES_A;
                    if (n2>n1)
                        return A_OR_B;
                }
                else if (p1.valExpr instanceof IntNE &&
                         p2.valExpr instanceof IntNE)
                {
                    if (n1==n2)
                        return EQUAL;
                    else
                        return A_OR_B;
                }
            }
        }

        // Default
        return TRUE;
    }

}
