/*
 * Used to construct and access prolog source files.
 */

package checkpolicy.prolog;

import java.io.*;
import java.util.*;

import jpl.*;
import jpl.fli.*;

public class PrologFile
{
    
    // Attributes
    private String filename = "";
    private StringBuilder code = new StringBuilder();

    // Accessors
    public String getFilename()
    {
        return filename;
    }
    
    // Constructor
    public PrologFile (String filename)
    {
        this.filename = filename;
    }
    
    // Methods
    public void appendCode(String newCode)
    {
        code.append(newCode);
    }
    
    /**
     * Creates the file anew, writing the code that has been given through
     * appendCode.
     * @throws java.io.IOException
     */
    public void writeCodeToFile() throws IOException
    {
        FileWriter out = new FileWriter(filename);
        out.write(code.toString());
        out.close();
    }

    /**
     * Initializes the Prolog engine. Recommended before opening new files.
     */
    public static void initPrologEngine()
    {
        JPL.init();
    }
    
    /**
     * Loads the specified Prolog file. Note that when creating new files
     * from scratch that you intend to use, you should first use
     * writeCodeToFile, then initPrologEngine, then loadPrologFile.
     */
    public void loadPrologFile()
    {
        Query q =
            new Query(
                "consult",
                new Term[] {new Atom(filename)}
            );
        if (q.hasSolution())
            System.out.println("Loaded " + filename + " successfully");
    }
    
    public void loadPrologFile(String file)
    {
        Query q =
            new Query(
                "consult",
                new Term[] {new Atom(file)}
            );
        if (q.hasSolution())
            System.out.println("Loaded " + file + " successfully");
    }

    public Hashtable[] runQuery(String queryString, boolean eliminateDuplicates)
    {
        Query query = new Query(queryString);
        
        if (!query.hasSolution())
            return null;
        
        Hashtable[] results = query.allSolutions();

        if (eliminateDuplicates)
            results = eliminateDuplicates(results);
        
        System.out.println(QueryBuilder.writeQuerySolutions(query,eliminateDuplicates));
        
        return results;
    }

    // TODO: FIX THIS
    private Hashtable[] eliminateDuplicates(Hashtable[] h)
    {
        HashSet<Hashtable> hSet = new HashSet<Hashtable>();
        for (int i = 0; i < h.length; i++)
        {
            hSet.add(h[i]);
        }
        Hashtable[] h2 = new Hashtable[hSet.size()];
        hSet.toArray(h2);
        return h2;
    }
    
    /**
     * Runs a query in Prolog and returns the processed result as a string. If
     * no results are found (i.e., we get a "false"), then null is returned.
     * @param queryString The query.
     * @param eliminateDuplicates Set to true if duplicate entries are unneeded.
     * @return The result of the query as a String.
     */
    public String queryToString(String queryString, boolean eliminateDuplicates)
    {
        Query query = new Query(queryString);
        
        if (!query.hasSolution())
            return null;
        
        return QueryBuilder.writeQuerySolutions(query, eliminateDuplicates);
    }
    
}
